//
// Created by zoe on 12/20/22.
//

#include "ConnectionManager.h"
#include "LTEConnectionException.hpp"
#include "ModemFunctionalityLevel.h"

#include <utility>

ConnectionManager::ConnectionManager(TinyGsm &client, GPRSCredentials credentials)
    : modem(client), credentials(std::move(credentials)) {}

void ConnectionManager::restartModem() {
    // Restart takes quite some time
    // To skip it, call init() instead of restart()
    esp_log_write(ESP_LOG_DEBUG, TAG_GSM.c_str(), "Restarting modem...\n");
    if (!modem.restart()) {
        esp_log_write(ESP_LOG_WARN, TAG_GSM.c_str(),
                      "Failed to restart modem, attempting to continue without restarting\n");
    }
}

void ConnectionManager::initModem() {
    esp_log_write(ESP_LOG_DEBUG, TAG_GSM.c_str(), "Initializing modem...\n");
    if (!modem.init()) {
        esp_log_write(ESP_LOG_WARN, TAG_GSM.c_str(),
                      "Failed to initialize modem, attempting to continue by restarting\n");
        // might not be the cleanest design
        restartModem();
    }
    setModemFunctionalityLevel(MINIMAL);
}

void ConnectionManager::setModemFunctionalityLevel(
    ModemFunctionalityLevel functionalityLevel) { // This sets the level of functionality of the modem. Full
                                                  // functionality is where the highest
    // level of power is drawn. Minimum functionality (default) is where the lowest level of power is drawn.
    // https://m2msupport.net/m2msupport/atcfun-set-phone-functionality/
    modem.sendAT(("+CFUN=" + std::to_string(ModemFunctionalityLevel::MINIMAL) + " ").c_str());
    if (modem.waitResponse(10000L) != 1) {
        DBG(" +CFUN=0  false ");
    }
}

void ConnectionManager::disableGPS() {
    // Set SIM7000G GPIO4 LOW ,turn off GPS power
    // CMD:AT+SGPIO=0,4,1,0
    // Only in version 20200415 is there a function to control GPS power
    modem.sendAT("+SGPIO=0,4,1,0");
    if (modem.waitResponse(10000L) != 1) {
        DBG(" SGPIO=0,4,1,0 false ");
    }
    modem.disableGPS();
}

void ConnectionManager::enableGPS() {
    // Set SIM7000G GPIO4 LOW ,turn on GPS power
    // CMD:AT+SGPIO=0,4,1,1
    // Only in version 20200415 is there a function to control GPS power
    modem.sendAT("+SGPIO=0,4,1,1");
    if (modem.waitResponse(10000L) != 1) {
        DBG(" SGPIO=0,4,1,1 false ");
    }
    modem.enableGPS();
}

void ConnectionManager::setNetworkMode(NetworkMode networkMode) {

    String res;
    res = modem.setNetworkMode(networkMode);
    if (res != "1") {
        DBG("setNetworkMode  false ");
        throw LTEConnectionException("Failed to set network mode");
    }
    delay(200);
}

void ConnectionManager::setPreferredMode(Mode mode) {
    /*AT+CBANDCFG=<mode>,<band>[,<band>…]
     * <mode> "CAT-M"   "NB-IOT"
     * <band>  The value of <band> must is in the band list of getting from  AT+CBANDCFG=?
     * For example, my SIM card carrier "NB-iot" supports B8.  I will configure +CBANDCFG= "Nb-iot ",8
     */
    modem.sendAT("+CBANDCFG=\"CAT-M\",8 ");
    if (modem.waitResponse(10000L) != 1) {
        DBG(" +CBANDCFG=\"NB-IOT\" ");
    }
    delay(200);
}
void ConnectionManager::setBand(BandMode bandMode, int band) {
    /*AT+CBANDCFG=<mode>,<band>[,<band>…]
     * <mode> "CAT-M"   "NB-IOT"
     * <band>  The value of <band> must is in the band list of getting from  AT+CBANDCFG=?
     * For example, my SIM card carrier "NB-iot" supports B8.  I will configure +CBANDCFG= "Nb-iot ",8
     */

    bandMode == BandMode::BAND_CAT_M ? modem.sendAT(("+CBANDCFG=\"CAT-M\"," + std::to_string(band) + " ").c_str())
                                     : modem.sendAT(("+CBANDCFG=\"NB-IOT\"," + std::to_string(band) + " ").c_str());

    if (modem.waitResponse(10000L) != 1) {
        DBG(" +CBANDCFG=\"NB-IOT\" ");
    }
    delay(200);
}
bool ConnectionManager::gprsConnect() {
    return modem.gprsConnect(credentials.apn.c_str(), credentials.user.c_str(), credentials.password.c_str());
}
bool ConnectionManager::isNetworkConnected() {
    return modem.isNetworkConnected();
}
std::string ConnectionManager::connect(const std::string &host, int port, RequestInformation requestInformation) {
    TinyGsmClient client{modem};
    if (!client.connect(host.c_str(), port)) {
        throw LTEConnectionException("Failed to connect to host");
    }

    String request = requestInformation.buildRequest();

    client.print(request);

    String line;
    // print response
    while (client.connected()) {
        line += client.readStringUntil('\n');
        if (line == "\r") {
            esp_log_write(ESP_LOG_DEBUG, TAG_GSM.c_str(), "headers received\n");
            break;
        }
    }
    client.stop();
    return line.c_str();
}
